from __future__ import annotations

from typing import Any


class TrackingIterator:
    def __init__(self, iterable) -> None:
        self.iterable = iter(iterable)
        self.count = 0

    def __iter__(self) -> TrackingIterator:
        return self

    def __next__(self) -> Any:
        self.count += 1
        return next(self.iterable)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestPrepend:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_prepend_hasNoItems(self,
                                iter_type,
                                iterable,
                                iterable_list):
        assert list(q(iterable).prepend(5)) == [5]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_prepend_hasSingleItem(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert list(q(iterable).prepend(1)) == [1] + iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_prepend_hasMultipleItems(self,
                                      iter_type,
                                      iterable,
                                      iterable_list):
        assert list(q(iterable).prepend(5)) == [5] + iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_prepend_hasMultipleItems_addingMultipleItems(self,
                                                          iter_type,
                                                          iterable,
                                                          iterable_list):
        assert list(q(iterable).prepend(1, 2, 3, 4)) == [1, 2, 3, 4] + iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_prepend_hasMultipleItems_addingSingleItemThatIsList(self,
                                                                 iter_type,
                                                                 iterable,
                                                                 iterable_list):
        assert list(q(iterable).prepend([1, 2, 3])) == [[1, 2, 3]] + iterable_list



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestAppendMany:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_appendMany_hasNoItems(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert list(q(iterable).append_many([1, 2, 3])) == [1, 2, 3]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_appendMany_hasSingleItem(self,
                                      iter_type,
                                      iterable,
                                      iterable_list):
        assert list(q(iterable).append_many([1, 2, 3])) == iterable_list + [1, 2, 3]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_appendMany_hasMultipleItems(self,
                                         iter_type,
                                         iterable,
                                         iterable_list):
        assert list(q(iterable).append_many([5, 6, 7])) == iterable_list + [5, 6, 7]



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestContains:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_contains_hasNoItems(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            assert not q(iterable).contains('0')
        else:
            assert not q(iterable).contains(0)

    def test_contains_hasSingleItem_itemIsContained(self):
        assert q([0]).contains(0)

    def test_contains_hasSingleItem_itemIsNotContained(self):
        assert not q([0]).contains(1)

    def test_contains_hasSingleNoneItem_nonesContained(self):
        assert q([None]).contains(None)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_contains_hasMultipleItems(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            assert q(iterable).contains('3')
        else:
            assert q(iterable).contains(3)



import csv
import timeit
from dataclasses import dataclass
from typing import Callable, Iterable


@dataclass
class NamedMethod:
    name: str
    method: Callable[[Iterable], None]


class BenchmarkRunner:
    def __init__(self, scenario_name: str, m1: NamedMethod, m2: NamedMethod,
                 dataset_generator: Callable[[int], Iterable]):
        self.scenario_name = scenario_name
        self.methods = [m1.method, m2.method]
        self.method_names = [m1.name, m2.name]
        self.dataset_generator = dataset_generator

    def _benchmark(self, method: Callable[[Iterable], None], dataset: Iterable):
        start_time = timeit.default_timer()
        method(dataset)
        return timeit.default_timer() - start_time

    def run(self, sizes: Iterable[int], output_csv: str):
        with open(output_csv, 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            header = ['Dataset'] + self.method_names
            csv_writer.writerow(header)

            for size in sizes:
                print(f"Running benchmark for dataset size: {size}")
                results = [size]
                for i, method in enumerate(self.methods):
                    print(f"Running benchmark for method: {self.method_names[i]}")
                    results.append(self._benchmark(method, self.dataset_generator(size)))
                csv_writer.writerow(results)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestOrderBy:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_orderBy_hasNoItems(self,
                                iter_type,
                                iterable,
                                iterable_list):
        assert list(q(iterable).order()) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_orderBy_hasSingleItem(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert list(q(iterable).order()) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_orderBy_hasMultipleItems_defaultSorting(self,
                                                     iter_type,
                                                     iterable,
                                                     iterable_list):
        assert list(q(iterable).order()) == list(iterable_list)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_orderBy_hasMultipleItems_customSelector(self,
                                                     iter_type,
                                                     iterable,
                                                     iterable_list):
        expected = list(sorted(iterable_list, key=lambda x: (int(x) % 2)))
        actual = list(q(iterable).order(by=lambda x: (int(x) % 2)))
        assert actual == expected



from dataclasses import dataclass
from functools import wraps
from typing import Optional


@dataclass
class Person:
    name: str
    age: int
    gender: str

    @staticmethod
    def default():
        return Person("No one", 0, "U")


def gen_person_name(num: int):
    # list of names for every alphabet letter
    names = ['anna', 'bob', 'charlie', 'david', 'emma', 'frank', 'gina', 'harry', 'ian', 'jane',
             'kate', 'larry',
             'mike', 'nancy', 'olivia', 'peter', 'quinn', 'robert', 'sarah', 'tom', 'ursula',
             'victor', 'william',
             'xavier', 'yvonne', 'zach']
    return names[num % 26] + str(num)


def gen_people(num: int):
    return [
        Person(
            name=gen_person_name(x),
            age=x,
            gender='M' if x % 2 == 0 else 'F'
        )
        for x in range(num)
    ]


@dataclass
class Point:
    x: int
    y: int

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)


class FliqTestUtils:
    @staticmethod
    def assertSmallerOrCloseTo(a: float,
                               b: float,
                               tolerance: float = 0.0,
                               msg: Optional[str] = None) -> None:
        """
        Asserts that a is smaller than or bigger but close to b.

        Parameters:
        - a, b: The two numbers to be compared.
        - tolerance: The tolerance to be used. Must be a float between 0 and 1.
        """
        if a <= b:
            return

        if not (0.0 <= tolerance <= 1.0):
            raise ValueError("Tolerance must be a float between 0 and 1.")

        # Calculate the allowed difference based on the tolerance
        max_diff = a * tolerance

        actual_diff = abs(a - b)

        assert actual_diff <= max_diff, (
            f"{a:.10f} and {b:.10f} "
            f"differ by more than {max_diff:.10f} ({tolerance * 100}%),"
            f" actual difference: {actual_diff:.10f} ({actual_diff / a * 100}%)"
            f" {msg if msg else ''}"
        )

    @classmethod
    def retry(cls, attempts: int):
        def decorator(fn):
            @wraps(fn)
            def wrapper(*args, **kwargs):
                for attempt in range(attempts):
                    try:
                        return fn(*args, attempt=attempt, **kwargs)
                    except AssertionError:
                        pass
                raise AssertionError(f"Test failed after {attempts} retries.")

            return wrapper

        return decorator


@dataclass
class MyTestClass:
    a: int
    b: int

    def __lt__(self, other):
        return self.a < other.a

    def __gt__(self, other):
        return self.a > other.a

    def __le__(self, other):
        return self.a <= other.a

    def __ge__(self, other):
        return self.a >= other.a

    def __eq__(self, other):
        return self.a == other.a

    def __add__(self, other):
        if not isinstance(other, MyTestClass):
            raise TypeError(f"Cannot add {type(self)} and {type(other)}")
        return MyTestClass(self.a + other.a, self.b + other.b)

    def __radd__(self, other):
        if isinstance(other, int):
            return MyTestClass(self.a + other, self.b + other)
        elif isinstance(other, MyTestClass):
            return MyTestClass(self.a + other.a, self.b + other.b)
        raise TypeError(f"Cannot add {type(self)} and {type(other)}")

    def __hash__(self):
        return hash(self.a)


class Params:
    sig_iterable = "iter_type,iterable,iterable_list"
    sig_iterable_obj = "iter_type,iterable"

    @staticmethod
    def iterable_obj_multi():
        return [
            ("list", [
                MyTestClass(0, 0 * 2),
                MyTestClass(1, 1 * 2),
                MyTestClass(2, 2 * 2),
                MyTestClass(3, 3 * 2),
                MyTestClass(4, 4 * 2)
            ]),
            ("generator", (MyTestClass(i, i * 2) for i in range(5))),
            ("tuple", (
                MyTestClass(0, 0 * 2),
                MyTestClass(1, 1 * 2),
                MyTestClass(2, 2 * 2),
                MyTestClass(3, 3 * 2),
                MyTestClass(4, 4 * 2))),
            ("set", {
                MyTestClass(0, 0 * 2),
                MyTestClass(1, 1 * 2),
                MyTestClass(2, 2 * 2),
                MyTestClass(3, 3 * 2),
                MyTestClass(4, 4 * 2)
            }),
        ]

    @staticmethod
    def iterable_obj_single():
        return [
            ("list", [MyTestClass(0, 0)]),
            ("generator", (MyTestClass(i, i) for i in range(1))),
            ("tuple", (MyTestClass(0, 0),)),
            ("set", {MyTestClass(0, 0)}),
        ]

    @staticmethod
    def iterable_multi_dup():
        """
        type_name, iterable, iterable_list
        """
        return [
            ("list", [0, 1, 2, 1, 3, 3, 4], [0, 1, 2, 1, 3, 3, 4]),
            ("generator", (i for i in [0, 1, 2, 1, 3, 3, 4]), [0, 1, 2, 1, 3, 3, 4]),
            ("tuple", (0, 1, 2, 1, 3, 3, 4), [0, 1, 2, 1, 3, 3, 4]),
            ("set", {0, 1, 2, 3, 4}, [0, 1, 2, 3, 4]),
            ("str", "0121334", ['0', '1', '2', '1', '3', '3', '4']),
        ]

    @staticmethod
    def iterable_multi():
        """
        type_name, iterable, iterable_list
        """
        return [
            ("list", [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]),
            ("generator", (i for i in range(5)), [0, 1, 2, 3, 4]),
            ("range", (range(5)), [0, 1, 2, 3, 4]),
            ("tuple", (0, 1, 2, 3, 4), [0, 1, 2, 3, 4]),
            ("set", {0, 1, 2, 3, 4}, [0, 1, 2, 3, 4]),
            ("str", "01234", ['0', '1', '2', '3', '4']),
        ]

    @staticmethod
    def iterable_single():
        """
        type_name, iterable, iterable_list
        """
        return [
            ("list", [0], [0]),
            ("generator", (i for i in range(1)), [0]),
            ("range", (range(1)), [0]),
            ("tuple", (0,), [0]),
            ("set", {0}, [0]),
            ("str", "0", ['0']),
        ]

    @staticmethod
    def iterable_empty():
        """
        type_name, actual, iterable_list
        """
        return [
            ("list", [], []),
            ("generator", (i for i in range(0)), []),
            ("range", (range(0)), []),
            ("tuple", (), []),
            ("set", set(), []),
            ("str", "", []),
        ]



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestEq:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_eq_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable) == iterable_list

    def test_eq_otherIsNotIterable_false(self):
        assert not q([0, 1, 2]) == 3

    def test_eq_listQueryToGenerator_equals(self):
        assert q([0, 1, 2]) == (range(3))

    def test_eq_generatorQueryToList_equals(self):
        assert q(range(3)) == [0, 1, 2]

    def test_eq_queryToQuery_equals(self):
        assert q([0, 1, 2]) == q([0, 1, 2])

    def test_eq_listQueryToGenerator_notEquals(self):
        assert q([0, 1, 2]) != (range(2))

    def test_eq_generatorQueryToList_notEquals(self):
        assert q(range(2)) != [0, 1, 2]

    def test_eq_queryToQuery_notEquals(self):
        assert q([0, 2, 1]) != q([0, 1, 2])

    def test_eq_subset_notEquals(self):
        assert q([0, 1]) != q([0, 1, 2])

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_eq_hasMultipleItems(self, iter_type, iterable, iterable_list):
        assert q(iterable) == iterable_list



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params, MyTestClass


class TestToDict:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_toDict_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).to_dict('id') == dict()

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_toDict_hasSingleItem_attributeDoesNotExist(self, iter_type, iterable):
        with pytest.raises(AttributeError):
            q(iterable).to_dict('id')

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_toDict_hasSingleItem_attributeExists_keyIsString(self, iter_type, iterable):
        assert q(iterable).to_dict('a') == {0: [MyTestClass(0, 0)]}

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_toDict_hasMultipleItems_attributeExists_keyIsString(self, iter_type, iterable):
        assert q(iterable).to_dict('a') == {
            0: [MyTestClass(0, 0)],
            1: [MyTestClass(1, 1 * 2)],
            2: [MyTestClass(2, 2 * 2)],
            3: [MyTestClass(3, 3 * 2)],
            4: [MyTestClass(4, 4 * 2)],
        }

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_toDict_hasMultipleItems_attributeExists_keyIsLambda(self, iter_type, iterable):
        assert q(iterable).to_dict(lambda x: x.a % 2 == 0) == {
            True: [MyTestClass(0, 0), MyTestClass(2, 2 * 2), MyTestClass(4, 4 * 2)],
            False: [MyTestClass(1, 1 * 2), MyTestClass(3, 3 * 2)],
        }



from unittest.mock import Mock

import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestFirstOrDefault:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_firstOrDefault_hasNoItems_defaultNotProvided(self,
                                                          iter_type,
                                                          iterable,
                                                          iterable_list):
        assert q(iterable).first_or_default() is None

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_firstOrDefault_hasNoItems_defaultProvided(self,
                                                       iter_type,
                                                       iterable,
                                                       iterable_list):
        some_obj = Mock()
        assert q(iterable).first_or_default(default=some_obj) == some_obj

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_firstOrDefault_hasSingleItem(self, iter_type, iterable, iterable_list):
        assert q(iterable).first_or_default() == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_firstOrDefault_hasMultipleItems_withoutFilter(self,
                                                           iter_type,
                                                           iterable,
                                                           iterable_list):
        assert q(iterable).first_or_default(default=10) == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_firstOrDefault_hasMultipleItems_withFilter(self,
                                                        iter_type,
                                                        iterable,
                                                        iterable_list):
        assert q(iterable).first_or_default(predicate=lambda x: int(x) > 10) is None



from time import perf_counter
from typing import Optional


class Timer:
    def __init__(self):
        self.elapsed: float = 0
        self.start: float = 0
        self.readout: Optional[str] = None

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.elapsed = perf_counter() - self.start
        self.readout = f'Time: {self.elapsed:.10f} seconds'



import random
from pathlib import Path
from typing import Iterable

import pandas as pd
import matplotlib.pyplot as plt

from benchmark_runner import NamedMethod, BenchmarkRunner
from csv_plotter import CsvPlotter
from fliq import q
from fliq.tests.fliq_test_utils import Person, gen_people


# s1: get the oldest male

def s1_fliq(dataset: Iterable):
    shuffled = list(dataset)
    random.shuffle(shuffled)
    return (q(dataset)
            .zip(shuffled)
            .where(lambda ps: ps[0].gender != ps[1].gender)
            .order(by=lambda ps: ps[0].age+ps[1].age)
            .take(5)
            .to_list())


def s1_std_lib(dataset: Iterable):
    shuffled = list(dataset)
    random.shuffle(shuffled)
    return sorted(
        filter(lambda ps: ps[0].gender != ps[1].gender, zip(dataset, shuffled)),
        key=lambda ps: ps[0].age+ps[1].age
    )[:5]


def s2_fliq(dataset: Iterable):
    first = [Person(name="Invitee1", age=i, gender='F') for i in range(len(dataset))]
    last = [Person(name="Invitee1", age=i, gender='F') for i in range(len(dataset))]
    return (q(dataset)
            .where(lambda p: 0 <= p.age < 100)
            .prepend_many(first)
            .append_many(last)
            .select(lambda p: p.name)
            .to_list())


def s2_std_lib(dataset: Iterable):
    first = [Person(name="Invitee1", age=i, gender='F') for i in range(len(dataset))]
    last = [Person(name="Invitee1", age=i, gender='F') for i in range(len(dataset))]
    return list(
        map(lambda p: p.name, first + list(
            filter(lambda p: 0 <= p.age < 100, dataset)) + last)
    )


def s3_fliq(dataset: Iterable):
    return q(dataset).select(lambda p: p.age * 2 if p.age % 2 == 0 else Person.default()).to_list()


def s3_std_lib(dataset: Iterable):
    return list((p.age * 2 if p.age % 2 == 0 else Person.default() for p in dataset))


def plot_benchmark(csv_path):
    # Read CSV file
    df = pd.read_csv(csv_path)

    # Rename x axis ticks to be the dataset size in thousands in the format of 1k, 10k, 1M
    def format_ticks(x):
        if x < 1000:
            return x
        elif x < 1_000_000:
            return f"{x // 1000}k"
        else:
            return f"{x // 1_000_000}M"
    df['Dataset'] = df['Dataset'].apply(format_ticks)

    # Set dataset as index
    df.set_index('Dataset', inplace=True)

    # Rename x axis to "Dataset Size"
    df.rename(columns={'Dataset': 'Dataset Size'}, inplace=True)

    # Plot the results in a compact way
    df.plot(kind='bar', figsize=(8, 4), rot=0, width=0.7)

    plt.title('Benchmark Results')
    plt.ylabel('Execution Time (seconds)')
    plt.xlabel('Dataset')

    plt.yscale('log')
    plt.grid(axis='y')
    plt.tight_layout()

    # add label to the top of each bar with a white background behind the black label
    for p in plt.gca().patches:
        plt.gca().annotate(f"{p.get_height():.3f}",
                           (p.get_x() + p.get_width() / 2, p.get_height()),
                           ha='center', va='center', color='b', xytext=(0, 10),
                           textcoords='offset points', bbox=dict(facecolor='white'))

    # save to the same filename as the csv file
    plt.savefig(csv_path.replace(".csv", ".png"))


BenchmarkRunner(
    scenario_name="Scenario 1",
    m1=NamedMethod("Fliq", s1_fliq),
    m2=NamedMethod("Standard Library", s1_std_lib),
    dataset_generator=gen_people,
).run(
    sizes=[
        1_00,
        10_000,
        1_000_000,
    ],
    output_csv="s1.csv",
)

BenchmarkRunner(
    scenario_name="Scenario 2",
    m1=NamedMethod("Fliq", s2_fliq),
    m2=NamedMethod("Standard Library", s2_std_lib),
    dataset_generator=gen_people,
).run(
    sizes=[
        1_00,
        10_000,
        1_000_000,
    ],
    output_csv="s2.csv",
)

BenchmarkRunner(
    scenario_name="Scenario 3",
    m1=NamedMethod("Fliq", s3_fliq),
    m2=NamedMethod("Standard Library", s3_std_lib),
    dataset_generator=gen_people,
).run(
    sizes=[
        1_00,
        10_000,
        1_000_000,
    ],
    output_csv="s3.csv",
)

CsvPlotter(Path('s1.csv')).plot_benchmark()
CsvPlotter(Path('s2.csv')).plot_benchmark()
CsvPlotter(Path('s3.csv')).plot_benchmark()





import pytest

from fliq.tests.fliq_test_utils import Params
from fliq import q


class TestIterations:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_forIterable_empty(self, iter_type, iterable, iterable_list):
        iterated = []
        for i in q(iterable):
            iterated.append(i)

        assert iterated == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_forIterable_single(self, iter_type, iterable, iterable_list):
        iterated = []
        for i in q(iterable):
            iterated.append(i)

        assert iterated == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_forIterable_multi(self, iter_type, iterable, iterable_list):
        iterated = []
        for i in q(iterable):
            iterated.append(i)

        assert iterated == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_stdIntegration_empty(self, iter_type, iterable, iterable_list):
        filtered = filter(lambda x: int(x) > 2, q(iterable))
        assert list(filtered) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_stdIntegration_single(self, iter_type, iterable, iterable_list):
        filtered = filter(lambda x: 2 > int(x), q(iterable))
        assert list(filtered) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_stdIntegration_multi(self, iter_type, iterable, iterable_list):
        filtered = filter(lambda x: int(x) > 2, q(iterable))
        assert list(filtered) == list(filter(lambda x: int(x) > 2, iterable_list))



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestRepr:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_repr_empty(self, iter_type, iterable, iterable_list):
        r = repr(q(iterable))
        assert r == "Query([])"

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_repr_single(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            return
        r = repr(q(iterable))
        assert r == "Query([0])"

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_repr_multiple(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            return
        r = repr(q(iterable))
        assert r == "Query([0, 1, 2, 3, 4])"

    def test_repr_biggerThanMaxRepresentation(self):
        r = repr(q(range(100)))
        assert (r == "Query([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ...])")



import ast
import re
from pathlib import Path
from typing import List


class MarkdownPythonEvaluator:
    def __init__(self, markdown_file_path: Path):
        self.markdown_file_path = markdown_file_path
        self.code_block_pattern = re.compile(r'```python(.*?)```', re.DOTALL)

    def extract_python_code_blocks(self) -> List[str]:
        """Extract python code blocks from the markdown file and split them into statements."""
        code_blocks = []
        with open(self.markdown_file_path, 'r', encoding='utf-8') as md_file:
            content = md_file.read()

            # Pattern to match python code blocks
            pattern = re.compile(r'```python(.*?)```', re.DOTALL)

            # Iterate over all matches
            for match in pattern.finditer(content):
                code_block = match.group(1).strip()  # Extract the code block
                code_blocks.append(code_block)

        return code_blocks

    def eval(self, code: str):
        """
        Safely evaluate a single python code block.
        This method does not allow modification of the environment or execution of arbitrary code.
        """
        try:
            # Safely evaluate expressions, preventing code execution
            block = ast.parse(code, mode='exec')
            exec(compile(block, '<string>', mode='exec'))
        except Exception as e:
            # In real use, you might want to handle the exception differently
            print(f"Error evaluating code block at {self.markdown_file_path}: "
                  f"{e} at block:\n{code}")
            exit(1)

    def evaluate_blocks(self):
        """Evaluate all extracted Python code blocks."""
        code_blocks = self.extract_python_code_blocks()
        for block in code_blocks:
            self.eval(block)


# Usage
root_path = Path(__file__).parent.parent
readme_path = root_path / 'README.md'
docs_path = root_path / 'docs'
files = [readme_path] + list(docs_path.glob('**/*.md'))
for f in files:
    print(f"Executing code blocks in {f}")
    evaluator = MarkdownPythonEvaluator(f)
    evaluator.evaluate_blocks()



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestZip:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_zip_hasNoItems(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).zip(range(5))) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_zip_hasSingleItem(self, iter_type, iterable, iterable_list):
        if iter_type == "str":
            return
        assert list(q(iterable).zip(range(5))) == [(0, 0)]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_zip_hasMultipleItems(self, iter_type, iterable, iterable_list):
        if iter_type == "str":
            return
        assert list(q(iterable).zip(range(5, 10))) == [
            (0, 5),
            (1, 6),
            (2, 7),
            (3, 8),
            (4, 9),
        ]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_zip_hasMultipleItems_multipleIterables(self, iter_type, iterable, iterable_list):
        if iter_type == "str":
            return
        assert list(q(iterable).zip(range(5, 10), range(10, 15))) == [
            (0, 5, 10),
            (1, 6, 11),
            (2, 7, 12),
            (3, 8, 13),
            (4, 9, 14),
        ]

    def test_zip_longest(self):
        iterable = [1, 2, 3]
        assert list(q(iterable).zip([4, 5], [6], longest=True)) == [
            (1, 4, 6),
            (2, 5, None),
            (3, None, None),
        ]

    def test_zip_longestWithFillValue(self):
        iterable = [1, 2, 3]
        fill_value = -1
        assert list(q(iterable).zip([4, 5], [6], longest=True, fillvalue=fill_value)) == [
            (1, 4, 6),
            (2, 5, -1),
            (3, -1, -1),
        ]

    def test_zip_shortestWithDifferentLengths(self):
        query = q(range(5))
        zipped = query.zip(range(3))
        assert list(zipped) == [(0, 0), (1, 1), (2, 2)]



from pathlib import Path
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import style


style.use('dark_background')  # Use a dark theme

class CsvPlotter:
    def __init__(self, csv_path: Path):
        self.csv_path = csv_path

    def plot_benchmark(self):
        df = pd.read_csv(self.csv_path)

        def format_ticks(x):
            if x < 1000:
                return x
            elif x < 1_000_000:
                return f"{x // 1000}k"
            else:
                return f"{x // 1_000_000}M"

        df['Dataset'] = df['Dataset'].apply(format_ticks)
        df.set_index('Dataset', inplace=True)
        df.rename(columns={'Dataset': 'Dataset Size'}, inplace=True)

        plt.rcParams["font.sans-serif"] = "Arial"  # Use a modern font
        plt.rcParams["font.size"] = 12  # Increase font size

        # Define Darcula-like colors for bars and labels
        bar_colors = ['#5294E2', '#E2777A']  # Soft blue and red
        label_color = '#FFC66D'  # Softer yellow for labels
        edge_color = '#646464'  # Gray for bar contour

        # Create the plot
        ax = df.plot(kind='bar', figsize=(8, 4), rot=0, width=0.5, color=bar_colors,
                     edgecolor=edge_color)

        # Set the background color
        ax.set_facecolor('#2B2B2B')

        # Set font
        plt.rcParams["font.family"] = "Helvetica"  # You can change this to your preferred font

        # Titles and labels with a lighter color for visibility
        plt.title('Benchmark Results', color='white')
        plt.ylabel('Execution Time (seconds)', color='white')
        plt.xlabel('Dataset', color='white')

        # Remove small ticks on the y-axis
        ax.tick_params(axis='y', which='minor', left=False)

        plt.yscale('log')
        plt.grid(axis='y', color='gray')  # Adjust grid color for visibility
        plt.tight_layout()

        # Add labels to bars
        for p in ax.patches:
            ax.annotate(f"{p.get_height():.3f}", (p.get_x() + p.get_width() / 2., p.get_height()),
                        ha='center', va='center', fontsize=9, color=label_color, xytext=(0, 10),
                        textcoords='offset points', bbox=dict(facecolor='black', edgecolor='0.1'))


        plt.savefig(str(self.csv_path.with_suffix('.png')), dpi=300)  # High-resolution output

# Usage remains the same




import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestPeek:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_peek_peekEmpty_fillvalueReturned(self, iter_type, iterable, iterable_list):
        sentinel = object()
        items = q(iterable)
        e0 = items.peek(fillvalue=sentinel)
        assert e0 is sentinel
        assert items == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_peek_peekSingle_doesNotConsume(self, iter_type, iterable, iterable_list):
        items = q(iterable)
        e0 = items.peek()
        assert e0 == iterable_list[0]
        assert items == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_peek_peekMoreThanSize_nonAvailableAreNone(self, iter_type, iterable, iterable_list):
        items = q(iterable)
        e0, e1 = items.peek(n=2)
        assert e0 == iterable_list[0]
        assert e1 is None
        assert items.to_list() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_peek_peekMultipleMoreThanSize_nonAvailableAreNone(self,
                                                               iter_type,
                                                               iterable,
                                                               iterable_list):
        items = q(iterable)
        e0, e1, e2, e3, e4 = items.peek(n=5)
        assert e0 == iterable_list[0]
        assert e1 is None
        assert e2 is None
        assert e3 is None
        assert e4 is None
        assert items.to_list() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_peek_peekedAtPopulatedQuery_doesNotConsume(self, iter_type, iterable, iterable_list):
        query = q(iterable)
        first_item = query.peek()
        assert first_item == iterable_list[0]
        assert query.to_list() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_peek_peekMultipleItems_doesNotConsume(self, iter_type, iterable, iterable_list):
        items = q(iterable)
        e0, e1, e2 = items.peek(n=3)
        assert e0 == iterable_list[0]
        assert e1 == iterable_list[1]
        assert e2 == iterable_list[2]
        assert items.to_list() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_peek_multiplePeekBeforeConsumption_doesNotConsume(self,
                                                               iter_type,
                                                               iterable,
                                                               iterable_list):
        items = q(iterable)
        e00, e01, e02 = items.peek(n=3)
        e10, e11, e12 = items.peek(n=3)
        assert e00 == iterable_list[0]
        assert e01 == iterable_list[1]
        assert e02 == iterable_list[2]
        assert e10 == iterable_list[0]
        assert e11 == iterable_list[1]
        assert e12 == iterable_list[2]
        assert items == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_peek_peekAfterPartialConsumption_doesNotConsume(self,
                                                             iter_type,
                                                             iterable,
                                                             iterable_list):
        items = q(iterable)
        e0 = next(items)
        e1, e2 = items.peek(n=2)
        assert e0 == iterable_list[0]
        assert e1 == iterable_list[1]
        assert e2 == iterable_list[2]
        assert items == iterable_list[1:]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_peek_peekDuringIteration_doesNotConsume(self,
                                                     iter_type,
                                                     iterable,
                                                     iterable_list):
        items = q(iterable)
        iterated = []
        peeked = None
        for i, element in enumerate(items):
            if i == 1:
                # peaked is item at index 2
                peeked = items.peek()
            iterated.append(element)
        assert iterated == iterable_list
        assert peeked == iterable_list[2]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_peek_peekAfterIteration_doesNotConsume(self,
                                                     iter_type,
                                                     iterable,
                                                     iterable_list):
        items = q(iterable)
        iterated = []
        for i, element in enumerate(items):
            iterated.append(element)
        peeked = items.peek()
        assert iterated == iterable_list
        assert peeked is None

    def test_peek_peekNonPositive_valueErrorRaised(self):
        with pytest.raises(ValueError):
            q(range(5)).peek(n=0)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params, MyTestClass


class TestBottom:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_bottom_hasNoItems(self,
                               iter_type,
                               iterable,
                               iterable_list):
        assert q(iterable).bottom(n=5) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_bottom_hasSingleItem(self,
                                  iter_type,
                                  iterable,
                                  iterable_list):
        assert q(iterable).bottom(n=5) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_bottom_hasMultipleItems_withoutSelector(self,
                                                     iter_type,
                                                     iterable,
                                                     iterable_list):
        if iter_type == "str":
            # not testing str because it requires a selector here, irrelevant for this case
            return
        assert q(iterable).bottom(n=3) == sorted(iterable_list, reverse=False)[:3]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_bottom_hasMultipleItems_withSelector(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        assert q(iterable).bottom(n=1, by=lambda x: -int(x)) == [max(iterable_list)]

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_bottom_hasMultipleObjectItems_withSelector(self,
                                                        iter_type,
                                                        iterable):
        assert q(iterable).bottom(n=2, by=lambda x: x.a) == [
            MyTestClass(0, 0),
            MyTestClass(1, 1 * 2)
        ]

    def test_bottom_hasMultipleItems_notEnoughItems(self):
        assert q(range(5)).bottom(n=100) == sorted(range(5), reverse=False)



import pytest

from fliq.tests.fliq_test_utils import Params
from fliq import q


class TestPartition:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_partition_hasNoItems(self,
                                  iter_type,
                                  iterable,
                                  iterable_list):
        q1, q2 = q(iterable).partition(lambda x: int(x), n=2)
        assert q1 == iterable_list
        assert q2 == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_partition_hasSingleItem(self,
                                     iter_type,
                                     iterable,
                                     iterable_list):
        q1, q2 = q(iterable).partition(lambda x: int(x), n=2)
        assert q1 == iterable_list
        assert q2 == []

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_partition_hasMultipleItems(self,
                                        iter_type,
                                        iterable,
                                        iterable_list):
        q0, q1, q2, q3, q4 = q(iterable).partition(lambda x: int(x), n=5)
        assert q0 == [iterable_list[0]]
        assert q1 == [iterable_list[1]]
        assert q2 == [iterable_list[2]]
        assert q3 == [iterable_list[3]]
        assert q4 == [iterable_list[4]]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_partition_hasMultipleItems_binary(self,
                                               iter_type,
                                               iterable,
                                               iterable_list):
        q0, q1 = q(iterable).partition(lambda x: int(x) % 2, n=2)
        assert q0 == [iterable_list[0], iterable_list[2], iterable_list[4]]
        assert q1 == [iterable_list[1], iterable_list[3]]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_partition_hasMultipleItems_predicate(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        q0, q1 = q(iterable).partition(lambda x: int(x) == 2)
        assert q0 == [iterable_list[0], iterable_list[1], iterable_list[3], iterable_list[4]]
        assert q1 == [iterable_list[2]]

    def test_partition_iteratingPastExhaustion_stopIterationRaised(self):
        q0, _ = q(range(3)).partition(lambda x: x % 2, n=2)
        next(q0)
        next(q0)
        with pytest.raises(StopIteration):
            next(q0)

    def test_partition_wrongPartitionIndexType_typeErrorRaised(self):
        q0, _, _ = q(range(3)).partition(lambda x: x == 1, n=3)
        with pytest.raises(TypeError):
            for _ in q0:
                pass

    def test_partition_iteratingOverItemWithoutPartition_valueErrorRaised(self):
        q0, _ = q(range(3)).partition(lambda x: x, n=2)
        next(q0)
        with pytest.raises(ValueError):
            next(q0)

    def test_partition_nonPositiveN_valueErrorRaised(self):
        with pytest.raises(ValueError):
            q(range(3)).partition(lambda x: x, n=0)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params, MyTestClass


class TestMin:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_min_hasNoItems_raisesException(self, iter_type, iterable, iterable_list):
        with pytest.raises(ValueError):
            q(iterable).min()

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_min_hasSingleItem(self, iter_type, iterable, iterable_list):
        assert q(iterable).min() == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_min_hasMultipleItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).min() == min(iterable_list)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_sum_hasMultipleItems_customSelector(self, iter_type, iterable, iterable_list):
        assert int(q(iterable).min(by=lambda x: int(x)*-1)) == 4

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_min_hasSingleItem_customObjects(self, iter_type, iterable):
        assert q(iterable).min() == MyTestClass(0, 0)

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_min_hasMultipleItems_customObjects(self, iter_type, iterable):
        assert q(iterable).min() == MyTestClass(0, 0)

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_min_hasSingleItem_customObjectsCustomAccumulator(self, iter_type, iterable):
        assert q(iterable).min(by=lambda o: o.a) == MyTestClass(0, 0)

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_min_hasMultipleItems_customObjectsCustomAccumulator(self, iter_type, iterable):
        assert q(iterable).min(by=lambda o: o.a) == MyTestClass(0, 0)




from __future__ import annotations

import collections.abc
import heapq
import itertools
import random
from collections import defaultdict
from functools import reduce
from itertools import islice, chain, zip_longest
from operator import attrgetter
from typing import Iterable, List, Optional, Any, Sized, Iterator, Callable, TYPE_CHECKING, Dict, \
    Union, Tuple, Hashable, Type, Generic, TypeVar, Sequence

from fliq.exceptions import (
    NoItemsFoundException, MultipleItemsFoundException, NotEnoughElementsException
)

if TYPE_CHECKING:
    from fliq import q  # noqa: F401 (used in docs)  # pragma: no cover


T = TypeVar('T')
U = TypeVar('U')

# editing of type aliases should also update api_intro.md
Predicate = Callable[[T], bool]
Selector = Callable[[T], U]

NumericSelector = Selector[T, Union[int, float]]
IndexSelector = Selector[T, int]


class Query(Generic[T], Iterable[T]):
    def __init__(self, iterable: Iterable[T]):
        """
        Create a Query object to allow fluent iterable processing
        """
        self._items: Iterable[T] = iterable
        self._iterator: Optional[Iterator[T]] = None

        # COW mode: copy-on-write mode, used to support snapshots
        self._cow_pending: bool = False

    def __iter__(self) -> Query[T]:
        if self._iterator is None:
            self._iterator = iter(self._items)
        return self

    def __next__(self) -> Union[T]:
        if self._iterator is None:
            self._iterator = iter(self._items)
        return next(self._iterator)

    def __contains__(self, item: Any) -> bool:
        return item in self._items

    def __eq__(self, other: Any) -> bool:
        """
        Compares the query to another iterable.

        Args:
            other: The iterable to compare to. Does not have to be a Query object.

        Returns:
            True if the query is equal to another iterable, item by item, ordered, False otherwise.
        """
        if not isinstance(other, Iterable):
            return False

        sentinel = object()  # To identify iterables of different sizes
        for a, b in zip_longest(self._items, other, fillvalue=sentinel):
            if a != b:
                # If two real objects returned unequal, return False
                # If one of the objects is the sentinel, iterables are of unequal size, return False
                return False

        return True

    def _self(self, updated_items: Optional[Iterable[U]] = None, in_snap: bool = False) -> Query[U]:
        """
        Adjusts the iterable of the query, and returns the query itself.
        This method abstract the need to create a new Query in some situations, to support
        the snapshot functionality.
        """
        updated_items = self._items if updated_items is None else updated_items  # type: ignore

        if in_snap:
            # COW prep: coming from snap, transform to list, don't copy object
            # This will allow multiple passes on this snapshot
            self._items = list(updated_items)  # type: ignore
            return self  # type: ignore
        elif self._cow_pending:
            # COW mode: coming from operation in COW mode, copy object
            # This will create a new query object in the first streamer following a snapshot
            # Following streamers will work on that same query object, minimizing object creations
            snapped_query = Query(iterable=updated_items)  # type: ignore
            return snapped_query
        else:
            # non-COW mode: coming from operation in non-COW mode, update iterable
            self._items = updated_items  # type: ignore
            return self  # type: ignore

    def __repr__(self) -> str:  # pragma: no cover
        """
        Returns a string representation of the query, while peeking at actual elements within
        the query. Query stays intact (i.e., no items are consumed).
        If representation is long, only the first 10 elements are shown, followed by an ellipsis.
        Elements are represented using their `repr` function.
        Inner iterable is represented as a list.

        Examples:
            >>> from fliq import q
            >>> q(range(10))
            Query([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
            >>> q(range(100))
            Query([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, ...])
            >>> q([])
            Query([])
            >>> repr(q(range(3)))
            'Query([0, 1, 2])'

        Returns:
            A string representation of the query.
        """
        repr_length = 10
        sentinel = object()
        elements = [
            repr(e) for e in
            self.peek(n=repr_length+1, fillvalue=sentinel) if e is not sentinel  # type: ignore
        ]
        peeked = ", ".join(elements[:repr_length])
        has_additional_items = len(elements) == repr_length+1
        if has_additional_items:
            peeked = peeked + ", ..."
        return f"Query([{peeked}])"

    # region Special Functionality

    def snap(self) -> Query[T]:
        """
        Yields the same elements, and creates a snapshot for the query.
        This snapshot allows for multiple iterations over the same "snapped" iterable.
        If multiple snapshots are created in a query lifetime, the last one is considered.

        Assumes a finite iterable.

        Examples:
            >>> from fliq import q
            >>> evens = q(range(10)).where(lambda x: x % 2 == 0).snap()
            >>> evens.where(lambda x: x > 3).count()
            3
            >>> evens.first()
            0
            >>> evens.select(lambda x: x ** 2).to_list()
            [0, 4, 16, 36, 64]
        """
        self._cow_pending = True
        return self._self(in_snap=True)

    def partition(self, by: Union[IndexSelector[T], Predicate[T]], n: int = 2) \
            -> Tuple[Query[T], ...]:
        # noinspection GrazieInspection
        """
        Yields n queries, each containing the elements that match the partition index selected.
        Supports infinite iterables,
        conditioned that there are finite sequences between elements of different partitions.

        Examples:
            >>> from fliq import q
            >>> first, second, third = q(range(10)).partition(lambda x: x % 3, n=3)
            >>> first.to_list(), second.to_list(), third.to_list()
            ([0, 3, 6, 9], [1, 4, 7], [2, 5, 8])
            >>> even, odd = q([1, 2, 3]).partition(lambda x: x % 2 == 0)
            >>> even.to_list(), odd.to_list()
            ([1, 3], [2])


        Args:
            by: IndexSelector that returns partition index for each element, in the range [0, n).
                Or, a Predicate to be used for a binary partition (when n=2).
                In case of a Selector, the first query will contain the elements in partition 0,
                the second query will contain the elements in partition 1, and so on.
                In case of a Predicate, the first query will contain the elements that don't satisfy
                 the predicate (for alignment between 0 and False as the first index).
            n: Optional. The number of partitions to create. Defaults to 2. Must be positive.
                When n=2, `by` can also be a Predicate.

        Raises:
            ValueError: In case the partition index is outside the range [0, n).
            TypeError: In case the partition index is not an integer.
        """
        if n <= 0:
            raise ValueError("Number of partitions must be positive")

        partition_queues: Dict[int, List[T]] = {i: [] for i in range(n)}

        def partition_generator(queue_partition_idx: int) -> Iterable[T]:
            for item in self:
                item_partition_idx: int = by(item)
                self._validate_partition_index(item_partition_idx, n)
                partition_queues[int(item_partition_idx)].append(item)
                if len(partition_queues[queue_partition_idx]) != 0:
                    # item found, stop generating and yield
                    yield partition_queues[queue_partition_idx].pop(0)

            # Iterator exhausted, drain the rest of the partition queue
            while len(partition_queues[queue_partition_idx]) != 0:
                yield partition_queues[queue_partition_idx].pop(0)

        # Create n Queries, each with its own partition generator
        return tuple(Query(partition_generator(i)) for i in range(n))

    def peek(self, n: int = 1, fillvalue: Any = None) -> Union[Optional[T], Sequence[Optional[T]]]:
        """
        Return the first n elements of the query, without exhausting the query.
        If n is 1, returns the first element as a single item, otherwise returns a tuple
        (that can be unpacked).
        If n is greater than the number of elements in the query,
        `fillvalue` will be returned for the remaining items (defaults to None).
        Use this to inspect the first n elements of the query, before consuming the query itself.
        Common use cases are logging and debugging.

        Examples:
            >>> from fliq import q
            >>> items = q(range(10))
            >>> i0, i1 = items.peek(2)
            >>> i0, i1
            (0, 1)
            >>> items.to_list()
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> a, b = q([]).peek(n=2)
            >>> a, b
            (None, None)

        Notes:
            Uses O(n) memory, as it materializes the first n elements.
            Supports infinite iterables.

        Args:
            n: Optional. The number of elements to peek. Defaults to 1.
            fillvalue: Optional. The value to use for padding when the query is exhausted.

        Raises:
            ValueError: In case n is not positive.

        Returns:
            A tuple of the first n elements of the query. If the query has less than n elements, the
                other missing elements will be returned as None.
        """
        if n < 1:
            raise ValueError(f"n must be positive, got {n}")

        # If we already have an iterator, use it; otherwise, get one from _items
        if self._iterator is None:
            self._iterator = iter(self._items)

        # Consume the first n items
        first_n_items: List[T] = list(islice(self._iterator, n))

        # If we consumed less than n items, pad with fillvalue
        padding: List[Optional[T]] = [fillvalue] * (n - len(first_n_items))

        # Adjust self._items to a new iterator that starts with the consumed items
        # followed by the remaining items
        self._items = chain(first_n_items, self._iterator)

        # Create a new iterator from the adjusted self._items
        self._iterator = iter(self._items)

        padded_result: Sequence[Optional[T]] = first_n_items + padding

        return padded_result if n > 1 else padded_result[0]

    # endregion

    # region Mappers

    def where(self, predicate: Optional[Predicate[T]] = None) -> Query[T]:
        """
        Yields elements that satisfy the predicate (aka filter).

        Examples:
            >>> from fliq import q
            >>> q(range(10)).where(lambda x: x % 2 == 0).to_list()
            [0, 2, 4, 6, 8]

        Args:
            predicate: Optional. The predicate to filter the query by. If None is
                given, no filtering takes place.
        """
        if predicate is None:
            # supported to ease syntax in higher level streamers and collectors
            return self._self()

        items = filter(predicate, self._items)
        return self._self(items)

    def select(self, selector: Selector[T, U]) -> Query[U]:
        """
        Yields the result of applying the selector function to each element (aka map).

        Examples:
            >>> from fliq import q
            >>> q(range(5)).select(lambda x: x * 2).to_list()
            [0, 2, 4, 6, 8]

        Args:
            selector: The selector function to apply to each element.
        """
        items = map(selector, self._items)
        return self._self(items)

    def exclude(self, predicate: Predicate[T]) -> Query[T]:
        """
        Yields elements that do not satisfy the predicate.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).exclude(lambda x: x > 3).to_list()
            [0, 1, 2, 3]

        Args:
            predicate: The predicate to filter the query by.
        """
        items = filter(lambda x: not predicate(x), self._items)
        return self._self(items)

    def distinct(self, preserve_order: bool = True) -> Query[T]:
        """
        Yields distinct elements.
        Distinct supports infinite iterables.
        Note that elements must be hashable.

        Examples:
            >>> from fliq import q
            >>> q([0, 1, 0, 2, 2]).distinct().to_list()
            [0, 1, 2]

        Raises:
            TypeError: In case one or more items in the query are not hashable.
            """

        def generator(internal_items: Iterable[T]) -> Iterable[T]:
            seen = set()
            for item in internal_items:
                if item not in seen:
                    seen.add(item)
                    yield item

        items = generator(self._items)

        return self._self(items)

    def order(self,
              by: Optional[Selector[T, U]] = None,
              ascending: bool = True) -> Query[T]:
        """Yields elements in sorted order.

        Examples:
            >>> from fliq import q
            >>> q([4, 3, 2, 1, 0]).order().to_list()
            [0, 1, 2, 3, 4]

        Args:
            by: a selector function to extract the key from an item, defaults to None.
                If None, the default ordering is used.
                If exists, assumes the key is comparable.
            ascending: whether to sort in ascending or descending order, defaults to True.
        """
        # assumed to be comparable
        if by is None:
            items = sorted(self._items, reverse=not ascending)  # type: ignore
        else:
            items = sorted(self._items, key=by, reverse=not ascending)  # type: ignore
        return self._self(items)

    def shuffle(self,
                buffer_size: int = 10,
                seed: Optional[Hashable] = None,
                fair: bool = False) -> Query[T]:
        """
        Yields elements in a random order.
        Supports infinite iterables.

        Args:
            buffer_size (int): The size of the shuffle buffer for an unfair shuffle. Defaults to 10.
                The bigger the buffer, the more memory is required,
                but the shuffle is closer to fair.
            seed: Optional. The seed to use for the random shuffle. Defaults to None.
            fair (bool): Whether to use a fair shuffle. Defaults to False.
                If True, each permutation of the elements will be equally likely.
                    This option does not support infinite iterables
                    (specifically, requires a sizeable iterable). This requires materialization
                    of the iterable, which is more memory intensive.
                    Use this for scenarios where fairness is important,
                    like rolling a die or shuffling a deck of cards.
                If False, each permutation of elements will NOT be equally likely.
                    However, this is more memory efficient, and supports infinite iterables.
                    The bigger the buffer size, the closer the shuffle will be to fair
                    (and thus, less memory efficient).
                    Use this for scenarios where perfect fairness is not important,
                    like shuffling a list of songs in a playlist.

        Examples:
            >>> from fliq import q
            >>> q(range(10)).shuffle(seed=42, buffer_size=5).to_list()
            [1, 5, 6, 3, 0, 8, 7, 2, 4, 9]
            >>> q(range(10)).shuffle(seed=42, fair=True).to_list()
            [7, 3, 2, 8, 5, 6, 9, 4, 0, 1]

        Raises:
            TypeError: In case a fair shuffle is requested for a non-sizeable iterable.
        """
        ran = random.Random(seed) if seed is not None else random.Random()

        if fair:
            if isinstance(self._items, List):
                shuffled_items = self._items
            elif isinstance(self._items, Sized):
                shuffled_items = list(self._items)
            else:
                raise TypeError("Fair shuffle is not supported for non-sizeable iterables")
            ran.shuffle(shuffled_items)
            return self._self(shuffled_items)

        def shuffled_generator(internal_items: Iterable[T]) -> Iterable[T]:
            buffer: List[T] = []
            sentinel = object()
            items_iter = iter(internal_items)

            # fill the buffer
            for _ in range(buffer_size):
                item = next(items_iter, sentinel)
                if item is sentinel:
                    # iterable exhausted (smaller than buffer size)
                    break
                buffer.append(item)  # type: ignore # (item is of type T)

            ran.shuffle(buffer)

            # if buffer was not filled, it means the iterable was too short
            if len(buffer) < buffer_size:
                yield from buffer
                return

            for item in items_iter:
                # yield a random item from the buffer and replace it with the new item
                idx = ran.randrange(len(buffer))
                yield buffer[idx]
                buffer[idx] = item

            # yield the remaining items in the buffer
            yield from buffer

        items = shuffled_generator(self._items)

        return self._self(items)

    def reverse(self) -> Query[T]:
        """
        Yields elements in reverse order.
        Notes:
         - in case of an irreversible query, TypeError is raised (e.g., set).
         - in case of an iterator, it is first converted to a list, then reversed,
         this has a performance and memory impact, and assumes a finite iterator.

         Examples:
             >>> from fliq import q
            >>> q([0, 1, 2, 3, 4]).reverse().to_list()
            [4, 3, 2, 1, 0]

        Raises:
            TypeError: In case the iterable is irreversible.
        """
        if isinstance(self._items, collections.abc.Iterator):
            items = reversed(list(self._items))
        else:
            items = reversed(self._items)  # type: ignore
        return self._self(items)

    def slice(self, start: int = 0, stop: Optional[int] = None, step: int = 1) -> Query[T]:
        """
        Yields a slice of the query

        Examples:
            >>> from fliq import q
            >>> q(range(10)).slice(start=1, stop=6, step=2).to_list()
            [1, 3, 5]

        Args:
            start: Optional. The start index of the slice. Defaults to 0.
            stop: Optional. The stop index of the slice. Defaults to None.
            step: Optional. The step of the slice. Defaults to 1.
        """
        items = islice(self._items, start, stop, step)
        return self._self(items)

    def take(self, n: int = 1, predicate: Optional[Predicate[T]] = None) -> Query[T]:
        """
        Yields up to n items that satisfy the predicate (if provided).
        In case the query is ordered, the first n elements are returned.

        Args:
            n: Optional. The number of elements to take. Defaults to 1.
            predicate: Optional. The predicate to filter the query by.
        """
        query = self.where(predicate).slice(stop=n)
        return self._self(query._items)

    def skip(self, n: int = 1) -> Query[T]:
        """
        Yields the elements after skipping the first n (as returned from the iterator).

        Examples:
            >>> from fliq import q
            >>> q(range(10)).skip(n=5).to_list()
            [5, 6, 7, 8, 9]

        Args:
            n: Optional. The number of items to take. Defaults to 1.
        """
        query = self.slice(start=n)
        return self._self(query._items)

    def zip(self, *iterables: Iterable[U], longest: bool = False, fillvalue: Optional[U] = None) \
            -> Query[Tuple[T, U]]:
        """
        Yields tuples of the elements of the query with the input iterables.
        The zipping stops as soon as the smallest of the iterables and the query is exhausted,
        unless longest is set to True, in which case the zipping stops when the longest iterable is
            exhausted. If strict mode is enabled, all iterables must have the same length.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).zip([5, 6, 7]).to_list()
            [(0, 5), (1, 6), (2, 7)]
            >>> q(range(5)).zip([5, 6, 7], longest=True).to_list()
            [(0, 5), (1, 6), (2, 7), (3, None), (4, None)]
            >>> q(range(5)).zip([5, 6, 7], [8, 9, 10]).to_list()
            [(0, 5, 8), (1, 6, 9), (2, 7, 10)]
            >>> q(range(5)).zip([5, 6, 7], [8, 9, 10], longest=True, fillvalue=-1).to_list()
            [(0, 5, 8), (1, 6, 9), (2, 7, 10), (3, -1, -1), (4, -1, -1)]

        Args:
            *iterables: One or more iterables to zip with the query.
            longest: If True, stop zipping when the longest iterable is exhausted.
                If False (default), stop when the shortest iterable is exhausted.
            fillvalue: The value to use for padding when the longest iterable is exhausted.
                relevant only when `longest` is True.
        """
        zip_func: Callable[..., Iterable[Tuple[T, U]]]
        if longest:
            zip_func = lambda *args: zip_longest(*args, fillvalue=fillvalue)  # noqa: E731
        else:
            zip_func = zip

        items: Iterable[Tuple[T, U]] = zip_func(self._items, *iterables)
        return self._self(items)

    def append(self, *single_items: T) -> Query[T]:
        """
        Yields the elements of the query, followed by the input element(s).
        API also supports multiple arguments, where each is considered as a single element.

        Infinite iterables are supported, behaving as expected.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).append(5).to_list()
            [0, 1, 2, 3, 4, 5]
            >>> q(range(5)).append(5, 6, 7).to_list()
            [0, 1, 2, 3, 4, 5, 6, 7]

        Args:
            *single_items: One or more elements to add to the end of the query.
        """
        items = chain(self._items, single_items)
        return self._self(items)

    def append_many(self, items: Iterable[T]) -> Query[T]:
        """
        Yields the elements of the iterable, followed by the elements given.

        Infinite iterables are supported, behaving as expected.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).append_many([5, 6, 7]).to_list()
            [0, 1, 2, 3, 4, 5, 6, 7]

        Args:
            items: An iterable to concatenate to the end of the query.

        Raises:
            TypeError: In case the elements are not iterable.
                Error will be raised when query is collected.
        """
        items = chain(self._items, items)
        return self._self(items)

    def prepend(self, *single_items: T) -> Query[T]:
        """
        Yields the element(s) given, followed by the elements of the query.
        API also supports multiple arguments, where each is considered as a single element.

        Infinite iterables are supported, behaving as expected.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).prepend(5).to_list()
            [5, 0, 1, 2, 3, 4]
            >>> q(range(5)).prepend(5, 6, 7).to_list()
            [5, 6, 7, 0, 1, 2, 3, 4]

        Args:
            *single_items: One or more elements to add to the start of the query.
        """
        items = chain(single_items, self._items)
        return self._self(items)

    def prepend_many(self, items: Iterable[T]) -> Query[T]:
        """
        Yields the elements given, followed by the elements of the query.

        Infinite iterables are supported, behaving as expected.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).prepend_many([5, 6, 7]).to_list()
            [5, 6, 7, 0, 1, 2, 3, 4]

        Args:
            items: The elements to add to the start of the query.

        Raises:
            TypeError: In case the items are not iterable.
                Error will be raised when the query is collected.
        """
        items = chain(items, self._items)
        return self._self(items)

    def group_by(self, key: Union[str, Selector[T, U]]) -> Query[List[T]]:
        """
        Yields an iterable of groups (as lists), where each group has identical key.
        If you require the group key, consider using `to_dict()` instead.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).group_by(lambda x: x % 2 == 0).to_list()
            [[1, 3], [2]]

        Args:
            key: A function that takes an element and returns its grouping key,
                or a string representing the name of an attribute to group by.
        """
        groups = self._group_by(key)

        return self._self(groups.values())

    def _group_by(self, key: Union[str, Selector[T, U]]) -> Dict[U, List[T]]:
        groups = defaultdict(list)
        key_selector: Selector[T, U]
        if callable(key):
            key_selector = key
        else:
            # key is a string
            key_selector = attrgetter(key)
        for item in self._items:
            group_key = key_selector(item)
            groups[group_key].append(item)
        return groups

    def top(self, n: int = 1, by: Optional[NumericSelector[T]] = None) -> Query[T]:
        """
        Yields the top n elements of the query, according to the selector (if provided).

        This is done in O(N log n) time, and O(n) space. Where n is the number of elements to take,
        and N is the number of elements in the query.

        Examples:
            >>> from fliq import q
            >>> q(range(10)).top(n=3).to_list()
            [9, 8, 7]
            >>> q(range(10)).top(n=3, by=lambda x: x*-1).to_list()
            [0, 1, 2]

        Args:
            n: Optional. The number of elements to take. Defaults to 1.
            by: Optional. The selector function to apply to each element. Defaults to the identity.
        """
        items: Iterable[T]
        if n <= 0:
            return self._self([])

        # heap holds tuples of (value, item), where value is either a numeric value
        # from the selector or the item itself if no selector is provided
        heap: List[Tuple[Union[T, int, float], T]] = []
        for item in self._items:
            value = item if by is None else by(item)
            if len(heap) < n:
                heapq.heappush(heap, (value, item))
            else:
                heapq.heappushpop(heap, (value, item))

        items = (
            Query(heap)
            .order(ascending=False)
            .select(lambda heap_pair: heap_pair[1])
        )

        return self._self(items)

    def bottom(self, n: int = 1, by: Optional[NumericSelector[T]] = None) -> Query[T]:
        """
        Yields the top n elements of the query, according to the selector (if provided).

        This is done in O(N log n) time, and O(n) space. Where n is the number of elements to take,
        and N is the number of elements in the query.

        Examples:
            >>> from fliq import q
            >>> q(range(10)).bottom(n=3).to_list()
            [0, 1, 2]

        Args:
            n: Optional. The number of elements to take. Defaults to 1.
            by: Optional. The selector function to apply to each element. Defaults to the identity.
                If by is None, the default ordering is used, which requires elements to be able to
                be multiplied by integers.
        """
        if by is None:
            return self.top(n=n, by=lambda x: -1 * x)  # type: ignore
        else:
            return self.top(n=n, by=lambda x: -1 * by(x))  # type: ignore # (by is not None)

    def flatten(
            self,
            max_depth: Optional[int] = None,
            ignore_types: Optional[Tuple[Type[Any], ...]] = (str, bytes)) -> Query[T]:
        """
        Yields a flattened iterable to a specified depth.

        Examples:
            >>> from fliq import q
            >>> q([[[1], 2], [3, 4]]).flatten().to_list()
            [1, 2, 3, 4]
            >>> q([[[1], 2], [3, 4]]).flatten(max_depth=1).to_list()
            [[1], 2, 3, 4]
            >>> q([['hello', 'world'], ['I', 'am', 'Fliq']]).flatten().to_list()
            ['hello', 'world', 'I', 'am', 'Fliq']
            >>> q([['hello', 'world'], ['I', 'am', 'Fliq']]).flatten(ignore_types=None).to_list()
            ['h', 'e', 'l', 'l', 'o', 'w', 'o', 'r', 'l', 'd', 'I', 'a', 'm', 'F', 'l', 'i', 'q']

        Args:
            max_depth: Optional. Non-negative. The maximum depth to flatten to. Defaults to none.
                (no limit, completely flat). If max_depth is 0, the query is unchanged.
            ignore_types: Optional. A tuple of types to ignore flattening for.
                Defaults to (str, bytes).
        Raises:
            ValueError: In case max_depth is non-positive.
        """
        if max_depth is not None and max_depth < 0:
            raise ValueError(f"max_depth must be non-negative (or -1 by default), got {max_depth}")

        def _flatten(iterable: Iterable[Any], current_depth: int) -> Iterable[Any]:
            for item in iterable:
                type_ignored = ignore_types is not None and isinstance(item, ignore_types)
                single_char_byte = isinstance(item, (str, bytes)) and len(item) == 1
                if not single_char_byte and isinstance(item, Iterable) and not type_ignored:
                    if max_depth is None or current_depth < max_depth:
                        yield from _flatten(item, current_depth + 1)
                    else:
                        yield item
                else:
                    yield item

        return self._self(_flatten(self._items, 0))

    def interleave(self, *iterables: Iterable[U]) -> Query[Union[T, U]]:
        """
        Combines the elements of the query with the elements of the provided iterables
        into a single sequence. The elements are interleaved in the order they appear in their
        respective sources (similar to zip). The process continues until all sources are exhausted.
        No padding is done in case the iterables are of different lengths.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).interleave([4], [5, 6]).to_list()
            [1, 4, 5, 2, 6, 3]

        Args:
            *iterables: One or more iterables to unify with the query.
        """
        sentinel = object()
        items: zip_longest[Tuple[T, U]] = zip_longest(self._items, *iterables,
                                                      fillvalue=sentinel)  # type: ignore
        flattened_items: Iterable[Union[T, U]] = chain.from_iterable(items)
        non_sentinel_items: Iterable[Union[T, U]] = filter(lambda x: x is not sentinel,
                                                           flattened_items)
        return self._self(non_sentinel_items)

    # endregion

    # region Materializers

    def first(self, predicate: Optional[Predicate[T]] = None) -> T:
        """
        Returns the first element in the query.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).first()
            1
            >>> q([]).first()
            Traceback (most recent call last):
            ...
            fliq.exceptions.NoItemsFoundException

        Args:
            predicate: Optional. The predicate to filter the query by.

        Raises:
            NoItemsFoundException: In case the query is empty.
        """
        query = self.where(predicate)
        try:
            return next(query)
        except StopIteration:
            raise NoItemsFoundException()

    def first_or_default(self,
                         predicate: Optional[Predicate[T]] = None,
                         default: Optional[T] = None) -> Optional[T]:
        """
        Returns the first element in the query, or a default value if the query is empty.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).first_or_default()
            1
            >>> q([]).first_or_default() # returns None

        Args:
            predicate: Optional. The predicate to filter the query by.
            default: Optional. The default value to return in case the query is empty.
                Defaults to None.
        """
        query = self.where(predicate)
        try:
            return next(query)
        except StopIteration:
            return default

    def single(self, predicate: Optional[Predicate[T]] = None) -> T:
        """
        Returns the single element in the query.

        Examples:
            >>> from fliq import q
            >>> q([1]).single()
            1
            >>> q([]).single()
            Traceback (most recent call last):
            ...
            fliq.exceptions.NoItemsFoundException

        Args:
            predicate: Optional. The predicate to filter the query by.

        Raises:
            NoItemsFoundException: In case the query is empty.
            MultipleItemsFoundException: In case the query has more than one element.
        """
        query = self.where(predicate)
        try:
            first = next(query)
        except StopIteration:
            raise NoItemsFoundException()

        try:
            second = next(query)
        except StopIteration:
            return first

        raise MultipleItemsFoundException(f"Found at least two items: {first}, {second}")

    def single_or_default(self,
                          predicate: Optional[Predicate[T]] = None,
                          default: Optional[T] = None) -> Optional[T]:
        """
        Returns the single element in the query, or a default value if the query is empty.

        Args:
            predicate: Optional. The predicate to filter the query by.
            default: Optional. The default value to return in case the query is empty.
                Defaults to None.

        Examples:
            >>> from fliq import q
            >>> q([1]).single_or_default()
            1
            >>> q([]).single_or_default() # returns None

            >>> q([1, 2, 3]).single_or_default()
            Traceback (most recent call last):
            ...
            fliq.exceptions.MultipleItemsFoundException

        Raises:
            MultipleItemsFoundException: In case the query has more than one element.
        """
        query = self.where(predicate)
        try:
            first = next(query)
        except StopIteration:
            return default

        try:
            next(query)
        except StopIteration:
            return first

        raise MultipleItemsFoundException()

    def sample(self,
               n: int = 1,
               seed: Optional[Hashable] = None,
               budget_factor: Optional[int] = 10,
               stop_factor: Optional[int] = 10) -> Union[T, List[T]]:
        """
        Returns a random sample of n elements from the query.
        If n is 1, returns a single item, otherwise returns a tuple (that can be unpacked).

        Examples:
            >>> from fliq import q
            >>> q(range(10)).sample(n=2, seed=42)
            [0, 4]
            >>> q(range(10)).sample(n=1, seed=42)
            1

        Args:
            n: Optional. The number of elements to sample. Defaults to 1.
            seed: Optional. The seed to use for the random sample. Defaults to None.
            budget_factor: Optional. Limits the number of attempts to sample n items,
                as a factor of n. Defaults to 10x n.
                None will disable budgeting, and may exhaust the iterable
                (depending on stop_factor).
            stop_factor: Optional. The probability to stop re-sampling once the sample is full,
                as a factor of n: 1 / ( stop_factor * n). Defaults to 1 / (10 * n).
                None will disable early stopping, and may exhaust the iterable
                (depending on budget_factor).

        Note:
            * To safely support infinite iterables, make sure you use
                budget_factor and/or stop_factor (they are set by default).

        Raises:
            NotEnoughItemsFoundException: In case the query is exhausted before n items are sampled.
            ValueError: In case n is not positive.
        """
        if n < 1:
            raise ValueError(f"n must be positive, got {n}")

        if budget_factor is not None and budget_factor < 1:
            raise ValueError(f"budget factor must be bigger than 1, got {budget_factor}")

        if stop_factor is not None and stop_factor < 1:
            raise ValueError(f"budget factor must be bigger than 1, got {stop_factor}")

        rand = random.Random(seed)

        reservoir: List[T] = []
        total_processed = 0
        budget = None if budget_factor is None else n * budget_factor
        stop_probability = None if stop_factor is None else 1 / (stop_factor * n)

        for element in itertools.islice(self._items, budget):
            total_processed += 1
            if len(reservoir) < n:
                reservoir.append(element)
            else:
                # Replace elements with decreasing probability
                replace_index = rand.randint(1, total_processed)
                if replace_index <= n:
                    reservoir[replace_index - 1] = element

                # Check for early stopping condition
                if stop_probability is not None and rand.random() < stop_probability:
                    break

        if len(reservoir) < n:
            raise NotEnoughElementsException("too many items requested")

        return reservoir[0] if n == 1 else reservoir

    def count(self) -> int:
        """
        Returns the number of elements in the query.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).count()
            3
        """
        # If the iterable is sized, return the length
        if isinstance(self._items, Sized):
            return len(self._items)

        # Otherwise, iterate over the iterable
        return sum(1 for _ in self)

    def any(self, predicate: Optional[Predicate[T]] = None) -> bool:
        """
        Returns whether any element in the query evaluates to true.
        If a predicate is provided, only elements that satisfy the predicate are considered.

        For custom types, consider providing a predicate or
          implementing `__bool__` or `__len__` to support this method.
         see https://docs.python.org/3/reference/datamodel.html#object.__bool__ .

        Examples:
            >>> from fliq import q
            >>> q([True, False, False]).any()
            True
            >>> q([False, False, False]).any()
            False

        Args:
            predicate: Optional. The predicate to filter the iterable by.
        """
        query = self.where(predicate)
        return any(query)

    def all(self, predicate: Optional[Predicate[T]] = None) -> bool:
        """
        Returns whether all elements in the query evaluate to true.
        If a predicate is provided, only elements that satisfy the predicate are considered.

        For custom types, consider providing a predicate or
          implementing `__bool__` or `__len__` to support this method.
         see https://docs.python.org/3/reference/datamodel.html#object.__bool__ .

        Examples:
            >>> from fliq import q
            >>> q([True, True, True]).all()
            True
            >>> q([True, False, True]).all()
            False

        Args:
            predicate: Optional. The predicate to filter the query by.
        """
        query = self.where(predicate)
        return all(query)

    def aggregate(self, by: Callable[[T, T], U], initial: Optional[U] = None) -> T:
        """
        Applies an accumulator function over the query.

        For an optimized summation of numeric values, use `sum`.

        Examples:
            >>> from fliq.tests.fliq_test_utils import Point
            >>> from fliq import q
            >>> q([Point(0, 0), Point(1, 1), Point(2, 2)]).aggregate(by=lambda a, b: a + b)
            Point(x=3, y=3)
            >>> q([Point(1, 1), Point(2, 2)]).aggregate(by=lambda a, b: a + b, initial=Point(0, 0))
            Point(x=3, y=3)

        Args:
            by: The accumulator function to apply to each two elements.
            initial: Optional. The initial value of the accumulator. Defaults to None.
                If provided, it will also serve as the default value for an empty query.
                If not provided, the first element of the query will be used as the initial value.
        """
        if initial is not None:
            return reduce(by, self._items, initial)  # type: ignore # (by items of the same type)
        else:
            return reduce(by, self._items)  # type: ignore # (by items of the same type)

    def max(self,
            by: Optional[Selector[T, U]] = None) -> T:
        """
        Returns the maximal element in the query.
        If a selector is provided, maximality is determined by the selected value.

        Custom types must provide a selector function or implement value comparisons
        (see https://docs.python.org/3/reference/expressions.html#value-comparisons).

        Examples:
            >>> from fliq import q
            >>> q(range(5)).max()
            4
            >>> q(range(5)).max(by=lambda x: x*-1)
            0

        Args:
            by: Optional. The selector function to test for the maximal element.
                If None, the default ordering is used.
                If exists, assumes the key is comparable.

        Raises:
            ValueError: In case the query is empty.
        """
        # assumed to be comparable
        if by is None:
            return max(self)  # type: ignore
        else:
            return max(self, key=by)  # type: ignore

    def min(self,
            by: Optional[Selector[T, U]] = None) -> T:
        """
        Returns the minimal element in the query.
        If a selector is provided, minimality is determined by the selected value.

        Custom types must provide a selector function or implement value comparisons
        (see https://docs.python.org/3/reference/expressions.html#value-comparisons).

        Examples:
            >>> from fliq import q
            >>> q(range(5)).min()
            0
            >>> q(range(5)).min(by=lambda x: x*-1)
            4

        Args:
            by: Optional. The selector function to test for the minimal element.
                If None, the default ordering is used.
                If exists, assumes the key is comparable.

        Raises:
            ValueError: In case the query is empty.
        """
        # assumed to be comparable
        if by is None:
            return min(self)  # type: ignore
        else:
            return min(self, key=by)  # type: ignore

    def contains(self, item: Any) -> bool:
        """
        Returns whether the query contains the given item (by equality, not identity).
        Query also supports the `in` and `not in` syntax, which are identical in functionality.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).contains(2)
            True
            >>> q([1, 2, 3]).contains(4)
            False
            >>> 2 in q([1, 2, 3])
            True

        Args:
            item: The item to test for.
        """
        return item in self._items

    def equals(self, other: Iterable[T], bag_compare: bool = False) -> bool:
        """
        Returns whether the query is equal to the given iterable.
        Query also supports the `==` and `!=` operators.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).equals([1, 2, 3])
            True
            >>> q([1, 2, 3]).equals(q([1, 2]))
            False
            >>> q([1, 2, 3]).equals([3, 2, 1], bag_compare=True)
            True

        Args:
            other: The iterable to test for equality.
            bag_compare: Optional. If True, compares the query and the other iterable as bags,
                ignoring order and duplicate items. Defaults to False.
        """
        if bag_compare:
            return set(self) == set(other)
        else:
            return self == other

    # region Numeric Collectors

    def sum(self,
            by: Optional[NumericSelector[T]] = None,
            accumulator: Union[int, float] = 0) -> Union[int, float]:
        """
        Returns the sum of the elements in the query.
        If a selector is provided, the sum of the selected elements is returned.
        If an accumulator is provided, it is used as the initial value for the summation.

        Custom types must provide a selector function or implement `__add__`
        and optionally `__radd__`
        (see https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types).

        Use this method for optimized summation of numeric values, for other types of aggregation,
         use aggregate.

        Examples:
            >>> from fliq import q
            >>> q(range(5)).sum()
            10
            >>> q(range(5)).sum(by=lambda x: x*2)
            20

        Args:
            by: Optional. The selector function to apply to each element.
            accumulator: Optional. The initial value of the sum. Defaults to 0.

        Returns:
            The sum of the elements in the iterable.
        """
        query = self
        if by is not None:
            query = self.select(by)  # type: ignore # (here `by` is subtype of select's `by`)
        return sum(query, start=accumulator)  # type: ignore # (query type is okay)

    # endregion

    def to_list(self) -> List[T]:
        """
        Returns the elements of the query as a list.
        """
        return list(self)

    def to_dict(self, key: Union[str, Selector[T, U]]) -> Dict[U, List[T]]:
        """
        Returns the elements of the query as a dictionary, grouped by the given key.
        If you don't require the group key, consider using `group_by()` instead.

        Examples:
            >>> from fliq import q
            >>> q([1, 2, 3]).to_dict(key=lambda x: x % 2 == 0)
            {False: [1, 3], True: [2]}

        Args:
            key: The selector function to apply to each element, or a string representing
                the name of an attribute to group by.
        """
        return dict(self._group_by(key))

    # endregion

    @staticmethod
    def _validate_partition_index(item_partition_idx: Union[int, bool], n: int) -> None:
        if isinstance(item_partition_idx, bool) and n != 2:
            raise TypeError(f"Partitioning by predicate is only supported for n=2, found n=`{n}`")
        if not (0 <= item_partition_idx < n):
            raise ValueError(f"Partition index {item_partition_idx} is not in range [0, {n})")



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestReverse:
    def _is_reversible(self, iter_type):
        return iter_type != "set"

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_reverse_reversible_empty(self,
                                      iter_type,
                                      iterable,
                                      iterable_list):
        if not self._is_reversible(iter_type):
            return

        assert list(q(iterable).reverse()) == list(reversed(iterable_list))

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_reverse_reversible_single(self,
                                       iter_type,
                                       iterable,
                                       iterable_list):
        if not self._is_reversible(iter_type):
            return

        assert list(q(iterable).reverse()) == list(reversed(iterable_list))

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_reverse_reversible_multi(self,
                                      iter_type,
                                      iterable,
                                      iterable_list):
        if not self._is_reversible(iter_type):
            return

        assert list(q(iterable).reverse()) == list(reversed(iterable_list))

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_reverse_irreversible_empty(self,
                                        iter_type,
                                        iterable,
                                        iterable_list):
        if self._is_reversible(iter_type):
            return

        with pytest.raises(TypeError):
            q(iterable).reverse()

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_reverse_irreversible_single(self,
                                         iter_type,
                                         iterable,
                                         iterable_list):
        if self._is_reversible(iter_type):
            return

        with pytest.raises(TypeError):
            q(iterable).reverse()

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_reverse_irreversible_multi(self,
                                        iter_type,
                                        iterable,
                                        iterable_list):
        if self._is_reversible(iter_type):
            return

        with pytest.raises(TypeError):
            q(iterable).reverse()



import pytest

from fliq.exceptions import NoItemsFoundException
from fliq.tests.fliq_test_utils import Params
from fliq import q


class TestSnap:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_noSnap_differentPasses_iterableExhausted(self, iter_type, iterable, iterable_list):
        evens = q(iterable).where(lambda x: int(x) % 2 == 0)
        evens.single(lambda x: int(x) > 2)
        with pytest.raises(NoItemsFoundException):
            evens.single(lambda x: int(x) < 0)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_snap_differentPasses_snapshotUsed(self, iter_type, iterable, iterable_list):
        evens = q(iterable).where(lambda x: int(x) % 2 == 0).snap()
        big = evens.single(lambda x: int(x) > 2)
        small = evens.single(lambda x: int(x) < 2)

        assert int(big) == 4
        assert int(small) == 0

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_snap_multipleSnaps_lastIsConsidered(self, iter_type, iterable, iterable_list):
        evens = (
            q(iterable)
            .where(lambda x: int(x) % 2 == 0)
            .snap()
            .select(lambda x: int(x) * 2)
            .snap()
        )
        first = evens.first(lambda x: x > 0)
        second = evens.first(lambda x: x > 0)

        assert first == 4
        assert second == 4

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_snap_snapNotLast(self, iter_type, iterable, iterable_list):
        evens = (
            q(iterable)
            .where(lambda x: int(x) % 2 == 0)
            .snap()
            .select(lambda x: int(x) * 2)
        )  # [0, 4, 8, 12, 16]
        first = evens.first()
        second = evens.first(lambda x: x > first)

        assert first == 0
        assert second == 4

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_noSnap_firstTwice_firstAndSecondReturned(self, iter_type, iterable, iterable_list):
        evens = (
            q((i for i in range(5)))
            .where(lambda x: int(x) % 2 == 0)
        )
        first1 = evens.first()
        first2 = evens.first()

        assert first1 == 0
        assert first2 == 2

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_snap_firstTwice_firstReturnedTwice(self, iter_type, iterable, iterable_list):
        evens = (
            q((i for i in range(5)))
            .where(lambda x: int(x) % 2 == 0)
            .snap()
        )
        first1 = evens.first()
        first2 = evens.first()

        assert first1 == 0
        assert first2 == 0

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_snap_streamerAndCollector(self, iter_type, iterable, iterable_list):
        evens = (
            q((i for i in range(5)))
            .where(lambda x: int(x) % 2 == 0)
            .snap()
        )  # [0, 2, 4]

        count = evens.count()
        first_doubled1 = evens.select(lambda x: x * 2).first()
        first_doubled2 = evens.select(lambda x: x * 2).first()

        assert count == 3
        assert first_doubled1 == 0
        assert first_doubled2 == 0



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestPrependMany:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_prependMany_hasNoItems(self,
                                    iter_type,
                                    iterable,
                                    iterable_list):
        assert list(q(iterable).prepend_many([5])) == [5]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_prependMany_hasSingleItem(self,
                                       iter_type,
                                       iterable,
                                       iterable_list):
        assert list(q(iterable).prepend_many([1, 2, 3])) == [1, 2, 3] + iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_prependMany_hasMultipleItems(self,
                                          iter_type,
                                          iterable,
                                          iterable_list):
        assert list(q(iterable).prepend_many([5, 6, 7])) == [5, 6, 7] + iterable_list



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestAny:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_any_hasNoItems(self, iter_type, iterable, iterable_list):
        assert not q(iterable).any()

    def test_any_hasSingleItem_integer(self):
        assert not q([0]).any()

    def test_any_hasSingleItem_nones(self):
        assert not q([None]).any()

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_any_hasMultipleItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).any()



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestInterleave:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_interleave_hasNoItems(self, iter_type, iterable, iterable_list):
        assert list(q(iterable).interleave(range(5))) == [0, 1, 2, 3, 4]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_interleave_hasSingleItem(self, iter_type, iterable, iterable_list):
        if iter_type == "str":
            return
        assert list(q(iterable).interleave(range(5))) == [0, 0, 1, 2, 3, 4]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_interleave_hasMultipleItems(self, iter_type, iterable, iterable_list):
        if iter_type == "str":
            return
        assert list(q(iterable).interleave(range(5, 10))) == [0, 5, 1, 6, 2, 7, 3, 8, 4, 9]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_interleave_hasMultipleItems_multipleIterables(self,
                                                           iter_type,
                                                           iterable,
                                                           iterable_list):
        if iter_type == "str":
            return
        assert list(q(iterable).interleave(range(5, 10), range(10, 15))) == \
            [0, 5, 10, 1, 6, 11, 2, 7, 12, 3, 8, 13, 4, 9, 14]

    def test_interleave_withDifferentLengths(self):
        query = q(range(5))
        zipped = query.interleave(range(3), range(1))
        assert zipped.to_list() == [0, 0, 0, 1, 1, 2, 2, 3, 4]



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestIn:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_in_hasNoItems(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            assert '0' not in q(iterable)
        else:
            assert 0 not in q(iterable)

    def test_contains_hasSingleItem_itemIsContained(self):
        assert 0 in q([0])

    def test_contains_hasSingleItem_itemIsNotContained(self):
        assert 1 not in q([0])

    def test_contains_hasSingleNoneItem_nonesContained(self):
        assert None in q([None])

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_contains_hasMultipleItems(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            assert '3' in q(iterable)
        else:
            assert 3 in q(iterable)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestAggregate:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_aggregate_hasNoItems_customInitial(self, iter_type, iterable, iterable_list):
        assert q(iterable).aggregate(by=lambda x, y: x+y, initial=0) == 0

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_aggregate_hasNoItems_noInitial_errorRaised(self, iter_type, iterable, iterable_list):
        with pytest.raises(TypeError):
            q(iterable).aggregate(by=lambda x, y: x+y)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_aggregate_hasSingleItem_noInitial(self, iter_type, iterable, iterable_list):
        assert q(iterable).aggregate(by=lambda x, y: x+y) == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_aggregate_hasMultipleItems(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            assert q(iterable).aggregate(by=lambda x, y: x+y) == ''.join(iterable_list)
        else:
            assert q(iterable).aggregate(by=lambda x, y: x+y) == sum(iterable_list)



class BaseQueryException(Exception):
    pass


class MultipleItemsFoundException(BaseQueryException):
    pass


class NoItemsFoundException(BaseQueryException):
    pass


class NotEnoughElementsException(BaseQueryException):
    pass



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestEquals:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_equals_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).equals(iterable_list)

    def test_equals_listQueryToGenerator_equals(self):
        assert q([0, 1, 2]).equals(range(3))

    def test_equals_generatorQueryToList_equals(self):
        assert q(range(3)).equals([0, 1, 2])

    def test_equals_queryToQuery_equals(self):
        assert q([0, 1, 2]).equals(q([0, 1, 2]))

    def test_eq_listQueryToGenerator_notEquals(self):
        assert not q([0, 1, 2]).equals(range(2))

    def test_eq_generatorQueryToList_notEquals(self):
        assert not q(range(2)).equals([0, 1, 2])

    def test_eq_queryToQuery_notEquals(self):
        assert not q([0, 2, 1]).equals(q([0, 1, 2]))

    def test_eq_bagCompare(self):
        assert q([0, 2, 1]).equals(q([0, 1, 2]), bag_compare=True)

    def test_eq_bagCompareWithDuplicates(self):
        assert q([0, 2, 1]).equals(q([0, 1, 2, 1]), bag_compare=True)

    def test_eq_subset_notEquals(self):
        assert not q([0, 1]).equals(q([0, 1, 2]))

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_eq_hasMultipleItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).equals(iterable_list)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestAppend:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_append_hasNoItems(self,
                               iter_type,
                               iterable,
                               iterable_list):
        assert list(q(iterable).append(5)) == [5]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_append_hasSingleItem(self,
                                  iter_type,
                                  iterable,
                                  iterable_list):
        assert list(q(iterable).append(1)) == iterable_list + [1]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_append_hasMultipleItems(self,
                                     iter_type,
                                     iterable,
                                     iterable_list):
        assert list(q(iterable).append(5)) == iterable_list + [5]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_append_hasMultipleItems_addingMultipleItems(self,
                                                       iter_type,
                                                       iterable,
                                                       iterable_list):
        assert list(q(iterable).append(1, 2, 3, 4)) == iterable_list + [1, 2, 3, 4]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_append_hasMultipleItems_addingSingleItemThatIsList(self,
                                                              iter_type,
                                                              iterable,
                                                              iterable_list):
        assert list(q(iterable).append([1, 2, 3])) == iterable_list + [[1, 2, 3]]



import pytest

from fliq import q
from fliq.exceptions import NotEnoughElementsException
from fliq.tests.fliq_test_utils import Params
from fliq.tests.utils.tracking_iterator import TrackingIterator


class TestSample:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_sample_sampleEmpty_errorRaised(self, iter_type, iterable, iterable_list):
        with pytest.raises(NotEnoughElementsException):
            q(iterable).sample()

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_sample_sampleSingle_doesNotConsume(self, iter_type, iterable, iterable_list):
        e0 = q(iterable).sample()
        assert e0 == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_sample_sampleMoreThanSize_errorRaised(self, iter_type, iterable, iterable_list):
        with pytest.raises(NotEnoughElementsException):
            q(iterable).sample(n=2)

    def test_sample_zeroNumber(self):
        with pytest.raises(ValueError):
            q(range(5)).sample(n=0)

    def test_sample_negativeNumber(self):
        with pytest.raises(ValueError):
            q(range(5)).sample(n=-1)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_sample_sampledOne_singleItemReturned(self, iter_type, iterable, iterable_list):
        sampled_item = q(iterable).sample(seed=42)
        assert sampled_item == iterable_list[1]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_sample_sampledWithinSize_multipleItemsReturned(self,
                                                            iter_type,
                                                            iterable,
                                                            iterable_list):
        s0, s1 = q(iterable).sample(n=2, seed=42)
        assert s0 == iterable_list[0]
        assert s1 == iterable_list[4]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_sample_sampledOverSize_errorRaised(self, iter_type, iterable, iterable_list):
        with pytest.raises(NotEnoughElementsException):
            q(iterable).sample(n=6)

    def test_sample_sampledWithBudget_multipleItemsReturned(self):
        n = 10
        budget_factor = 1
        sequence = range(100)  # Known sequence
        tracking_iterable = TrackingIterator(sequence)
        q(tracking_iterable).sample(n=n, budget_factor=budget_factor)
        assert tracking_iterable.count <= n * budget_factor

    def test_sample_sampleWithStopFactor_stopFactorStopsSampling(self):
        sequence = range(10000)  # Longer sequence to allow for early stopping
        n = 5
        seed = 42
        stop_factor = 1  # Higher probability of stopping early
        budget_factor = 100
        tracking_iterable = TrackingIterator(sequence)

        q(tracking_iterable).sample(n=n,
                                    seed=seed,
                                    budget_factor=budget_factor,
                                    stop_factor=stop_factor)

        assert tracking_iterable.count <= budget_factor * n

    def test_sample_budgetFactorIsPositive(self):
        with pytest.raises(ValueError):
            q(range(5)).sample(budget_factor=-1)

    def test_sample_stopFactorIsPositive(self):
        with pytest.raises(ValueError):
            q(range(5)).sample(stop_factor=-1)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params, MyTestClass


class TestTop:
    def test_top_topZero(self):
        assert q(range(10)).top(n=0) == []

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_top_hasNoItems(self,
                            iter_type,
                            iterable,
                            iterable_list):
        assert q(iterable).top(n=5) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_top_hasSingleItem(self,
                               iter_type,
                               iterable,
                               iterable_list):
        assert q(iterable).top(n=5) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_top_hasMultipleItems_withoutSelector(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        assert q(iterable).top(n=3) == sorted(iterable_list, reverse=True)[:3]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_top_hasMultipleItems_withSelector(self,
                                               iter_type,
                                               iterable,
                                               iterable_list):
        assert q(iterable).top(n=1, by=lambda x: -int(x)) == [min(iterable_list)]

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_top_hasMultipleObjectItems_withSelector(self,
                                                     iter_type,
                                                     iterable):
        assert q(iterable).top(n=2, by=lambda x: x.a) == [
            MyTestClass(4, 4 * 2),
            MyTestClass(3, 3 * 2)
        ]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_top_hasMultipleItems_notEnoughItems(self,
                                                 iter_type,
                                                 iterable,
                                                 iterable_list):
        assert q(iterable).top(n=100) == sorted(iterable_list, reverse=True)



from dataclasses import dataclass
from unittest import TestCase

from fliq import q
from fliq.tests.fliq_test_utils import Person


@dataclass
class Item:
    price: int
    rating: float


class TestSanity(TestCase):

    def setUp(self):
        self.items = (Item(i, i % 5) for i in range(100))

    def test_sanity_1(self):
        price = (
            q(self.items)
            .where(lambda x: x.price > 50)
            .select(lambda x: x.price)
            .first_or_default(default=0)
        )
        assert price == 51

    def test_sanity_2(self):
        def name_selector(person: Person) -> str:
            return person.name

        men = [Person(name=f'Mr {i}', age=i, gender="M") for i in range(30)]
        women = [Person(name=f'Ms {i+10}', age=i, gender="W") for i in range(30)]
        non_binary = [Person(name=f'{i+30}', age=i, gender="N") for i in range(30)]
        adults = (q(men)
                  .append_many(women)
                  .where(lambda p: p.age >= 25)
                  .prepend_many(non_binary)
                  .select(name_selector))

        assert adults.count() == 5 + 5 + 30

    def test_sanity_3(self):
        data = [1, 2, 3, 4, 5]
        gen = (i * 2 if i % 2 == 0 else -1 for i in [1, 2, 3, 4, 5])
        fliq = q(data).select(lambda x: x * 2 if x % 2 == 0 else -1)
        assert list(fliq) == list(gen)


import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params, MyTestClass


class TestSum:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_sum_hasNoItems_customAccumulator(self, iter_type, iterable, iterable_list):
        assert q(iterable).sum(accumulator=1) == 1

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_sum_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).sum() == 0

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_sum_hasSingleItem(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            # sum should be used with numbers only
            return
        assert q(iterable).sum() == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_sum_hasMultipleItems(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            # sum should be used with numbers only
            return
        assert q(iterable).sum() == sum(iterable_list)

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_sum_hasMultipleItems_customSelector(self, iter_type, iterable, iterable_list):
        if iter_type == 'str':
            # sum should be used with numbers only
            return
        assert q(iterable).sum(by=lambda x: x*2) == sum([x*2 for x in iterable_list])

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_sum_hasSingleItem_customObjects(self, iter_type, iterable):
        assert q(iterable).sum(by=lambda o: o.a) == 0

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_sum_hasMultipleItems_customObjects(self, iter_type, iterable):
        assert q(iterable).sum(by=lambda o: o.a) == 10

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_sum_hasSingleItem_customObjects_customAccumulator(self, iter_type, iterable):
        assert q(iterable).sum(accumulator=MyTestClass(0, 0)) == MyTestClass(0, 0)

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_sum_hasMultipleItems_customObjects_customAccumulator(self, iter_type, iterable):
        assert q(iterable).sum(accumulator=MyTestClass(0, 0)) == MyTestClass(10, 10 * 2)

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_sum_hasMultipleItems_customObjectsNoAccumulator(self, iter_type, iterable):
        assert q(iterable).sum() == MyTestClass(10, 10 * 2)



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestSkip:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_skip_hasNoItems(self,
                             iter_type,
                             iterable,
                             iterable_list):
        assert list(q(iterable).skip(n=5)) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_skip_hasSingleItem_takeBiggerThanIterableSize(self,
                                                           iter_type,
                                                           iterable,
                                                           iterable_list):
        assert list(q(iterable).skip(n=5)) == []

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_take_hasMultipleItems(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert list(q(iterable).skip(n=3)) == iterable_list[3:]



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params, MyTestClass


class TestGroupBy:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_groupBy_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).group_by('id') == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_groupBy_hasSingleItem_attributeDoesNotExist(self, iter_type, iterable):
        with pytest.raises(AttributeError):
            q(iterable).group_by('id')

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_single())
    def test_groupBy_hasSingleItem_attributeExists_keyIsString(self, iter_type, iterable):
        assert q(iterable).group_by('a') == [[MyTestClass(0, 0)]]

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_groupBy_hasMultipleItems_attributeExists_keyIsString(self, iter_type, iterable):
        assert q(iterable).group_by('a') == [
            [MyTestClass(0, 0)],
            [MyTestClass(1, 1 * 2)],
            [MyTestClass(2, 2 * 2)],
            [MyTestClass(3, 3 * 2)],
            [MyTestClass(4, 4 * 2)],
        ]

    @pytest.mark.parametrize(Params.sig_iterable_obj, Params.iterable_obj_multi())
    def test_groupBy_hasMultipleItems_attributeExists_keyIsLambda(self, iter_type, iterable):
        assert q(iterable).group_by(lambda x: x.a % 2 == 0) == [
            [MyTestClass(0, 0), MyTestClass(2, 2 * 2), MyTestClass(4, 4 * 2)],
            [MyTestClass(1, 1 * 2), MyTestClass(3, 3 * 2)],
        ]



import random

import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params

SEED = 42


@pytest.fixture()
def rand() -> random.Random:
    return random.Random(SEED)


class TestShuffle:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_shuffle_hasNoItems_fairShuffle(self,
                                            iter_type,
                                            iterable,
                                            iterable_list):
        if iter_type == 'generator':
            with pytest.raises(TypeError):
                q(iterable).shuffle(fair=True)
        else:
            assert q(iterable).shuffle(fair=True) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_shuffle_hasNoItems_unfairShuffle(self,
                                              iter_type,
                                              iterable,
                                              iterable_list):
        assert q(iterable).shuffle(fair=False) == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_shuffle_hasSingleItem(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert q(iterable).shuffle() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_shuffle_hasMultipleItems(self,
                                      iter_type,
                                      iterable,
                                      iterable_list,
                                      rand):
        rand.shuffle(iterable_list)
        assert q(iterable).shuffle(seed=SEED) == iterable_list

    def test_shuffle_hasMultipleItems_unfairShuffle_bufferSmallerThanIterableSize(self):
        shuffled_items = list(range(100))
        unfair_shuffled_items = q(range(100)).shuffle(seed=SEED, buffer_size=10).to_list()
        assert set(unfair_shuffled_items) == set(shuffled_items)
        assert len(unfair_shuffled_items) == len(shuffled_items)

    def test_shuffle_hasMultipleItems_unfairShuffle_bufferBiggerThanIterableSize(self):
        shuffled_items = list(range(10))
        unfair_shuffled_items = q(range(10)).shuffle(seed=SEED, buffer_size=100).to_list()
        assert set(unfair_shuffled_items) == set(shuffled_items)
        assert len(unfair_shuffled_items) == len(shuffled_items)

    def test_shuffle_hasMultipleItems_unfairShuffle_bufferSameSizeAsIterableSize(self):
        shuffled_items = list(range(100))
        unfair_shuffled_items = q(range(100)).shuffle(seed=SEED, buffer_size=100).to_list()
        assert set(unfair_shuffled_items) == set(shuffled_items)
        assert len(unfair_shuffled_items) == len(shuffled_items)

    def test_shuffle_hasMultipleItems_fairShuffle(self, rand):
        shuffled_items = list(range(100))
        rand.shuffle(shuffled_items)
        assert q(range(100)).shuffle(seed=SEED, fair=True) == shuffled_items



import pytest

from fliq import q
from fliq.exceptions import NoItemsFoundException
from fliq.tests.fliq_test_utils import Params


class TestFirst:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_first_hasNoItems(self,
                              iter_type,
                              iterable,
                              iterable_list):
        with pytest.raises(NoItemsFoundException):
            q(iterable).first()

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_first_hasSingleItem(self,
                                 iter_type,
                                 iterable,
                                 iterable_list):
        assert q(iterable).first() == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_first_hasMultipleItems_withoutFilter(self,
                                                  iter_type,
                                                  iterable,
                                                  iterable_list):
        assert q(iterable).first() == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_first_hasMultipleItems_withFilter(self,
                                               iter_type,
                                               iterable,
                                               iterable_list):
        assert q(iterable).first(lambda x: int(x) > 0) == iterable_list[1]



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestAll:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_all_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).all()

    def test_all_hasSingleItem_integer(self):
        assert q([1]).all()

    def test_all_hasSingleItem_nones(self):
        assert not q([None]).all()

    def test_all_hasMultipleItems(self):
        assert q((1, 2, 3, True)).all()



import pytest

from fliq import q
from fliq.tests.fliq_test_utils import Params


class TestFlatten:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_flatten_hasNoItems(self,
                                iter_type,
                                iterable,
                                iterable_list):
        assert q(iterable).flatten() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_flatten_hasSingleItem(self,
                                   iter_type,
                                   iterable,
                                   iterable_list):
        assert q(iterable).flatten() == iterable_list

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_flatten_hasMultipleItems(self,
                                      iter_type,
                                      iterable,
                                      iterable_list):
        assert q(iterable).flatten() == iterable_list

    def test_flatten_nestedTwoLevels_unlimited(self):
        assert q([[1, 2], [3, 4]]).flatten() == [1, 2, 3, 4]

    def test_flatten_nestedThreeLevels_unlimited(self):
        assert q([[[1, 2]], [[3, 4]]]).flatten() == [1, 2, 3, 4]

    def test_flatten_nestedTwoLevels_limited(self):
        assert q([[1, 2], [3, 4]]).flatten(max_depth=2) == [1, 2, 3, 4]

    def test_flatten_nestedThreeLevels_limited(self):
        assert q([[[1, 2]], [[3, 4]]]).flatten(max_depth=1).to_list() == [[1, 2], [3, 4]]

    def test_flatten_nestedThreeLevels_limitedMoreThanExists(self):
        assert q([[[1, 2]], [[3, 4]]]).flatten(max_depth=10) == [1, 2, 3, 4]

    def test_flatten_nestedString_ignoredByDefault(self):
        assert q([['hello'], 'world']).flatten(max_depth=3) == ['hello', 'world']

    def test_flatten_nestedString_notIgnoredWithLimited(self):
        assert (q([['hello'], 'world']).flatten(max_depth=2, ignore_types=None).to_list() ==
                ['h', 'e', 'l', 'l', 'o', 'w', 'o', 'r', 'l', 'd'])

    def test_flatten_nestedString_notIgnoredUnlimited(self):
        assert (q([['hello', 'world'], ['I', 'am', 'Fliq']]).flatten(ignore_types=None).to_list() ==
                ['h', 'e', 'l', 'l', 'o', 'w', 'o', 'r',
                 'l', 'd', 'I', 'a', 'm', 'F', 'l', 'i', 'q'])

    def test_flatten_nestedStringAsGenerator_notIgnoredUnlimited(self):
        assert (q(str(i) + str(i + 1) for i in range(5)).flatten(ignore_types=None).to_list() ==
                ['0', '1', '1', '2', '2', '3', '3', '4', '4', '5'])

    def test_flatten_negativeDepth_valueErrorRaised(self):
        with pytest.raises(ValueError):
            q([1, 2, 3]).flatten(max_depth=-1)


import pytest

from fliq import q
from fliq.exceptions import MultipleItemsFoundException
from fliq.tests.fliq_test_utils import Params


class TestSingleOrDefault:
    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_empty())
    def test_singleOrDefault_hasNoItems(self, iter_type, iterable, iterable_list):
        assert q(iterable).single_or_default() is None

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_single())
    def test_singleOrDefault_hasSingleItem(self, iter_type, iterable, iterable_list):
        assert q(iterable).single_or_default() == iterable_list[0]

    @pytest.mark.parametrize(Params.sig_iterable, Params.iterable_multi())
    def test_singleOrDefault_hasMultipleItems(self, iter_type, iterable, iterable_list):
        with pytest.raises(MultipleItemsFoundException):
            q(iterable).single_or_default()


import inspect
import textwrap
from pathlib import Path

from fliq import Query


docs_path = Path(__file__).parent.parent / 'docs'


def get_md_body(title, methods):
    template = """
    # {{title}}
    
    ::: fliq.query.Query
        options:
            filters: [
                {{filters}} 
            ]   
    """
    body = (template
            .replace('{{title}}', title)
            .replace('{{filters}}', ', '.join(methods)))
    return textwrap.dedent(body)


def create_api_docs():
    docs_reference_path = docs_path / 'reference' / 'code_api'
    mappers = []
    materializers = []

    # skipped methods will not be included in the mapper methods and the materializer methods
    # assumed to have their own documentation page or not relevant to the user
    skipped_methods = [
        Query.__init__.__name__,
        Query._self.__name__,
        Query.__iter__.__name__,
        Query.__next__.__name__,
        Query.__repr__.__name__,
        Query.__eq__.__name__,
        Query.snap.__name__,
        Query.peek.__name__,
        Query.partition.__name__,
    ]

    for name, method in inspect.getmembers(Query, predicate=inspect.isfunction):
        if name in skipped_methods:
            # skip constructor
            continue
        ret_type = inspect.signature(method).return_annotation
        is_streamer = ret_type.startswith('Query')
        exact_name_regex = f'"^{name}$"'
        if is_streamer:
            mappers.append(exact_name_regex)
        else:
            materializers.append(exact_name_regex)

    with open(docs_reference_path / 'mapper_methods.md', 'w') as f:
        f.write(get_md_body('Mapper Methods', mappers))

    with open(docs_reference_path / 'materializer_methods.md', 'w') as f:
        f.write(get_md_body('Materializer Methods', materializers))


def create_index_doc():
    # this function copies README.md to docs/index.md
    index_doc_path = docs_path / 'index.md'
    readme_path = Path(__file__).parent.parent / 'README.md'

    with open(index_doc_path, 'w') as f:
        with open(readme_path, 'r') as readme:
            # update relative path for asserts
            content = readme.read().replace(
                "docs/assets/",
                "assets/"
            )
            f.write(content)


create_api_docs()
create_index_doc()